from typing import Callable, Tuple, List


class DifferentialEquation:
    def __init__(
        self,
        f: Callable,
        iterations: int,
        h: float,
        x: float,
        y: float,
    ) -> None:
        self.f = f
        self.iterations = iterations
        self.h = h
        self.x = x
        self.y = y

    def euler(self) -> Tuple[float, float, List[Tuple[float, float]]]:
        history = [(self.x, self.y)]
        for i in range(self.iterations):
            self.y += self.h * self.f(self.x, self.y)
            self.x += self.h
            history.append((self.x, self.y))
        return (self.x, self.y), history

    def advanced_euler(self) -> Tuple[float, float]:
        history = [(self.x, self.y)]
        for i in range(self.iterations):
            f_res = self.f(self.x, self.y)
            y_temp = self.y + self.h * f_res
            f_temp = self.f(self.x + self.h, y_temp)
            self.y += (f_res + f_temp) / 2 * self.h
            self.x += self.h
            history.append((self.x, self.y))
        return (self.x, self.y), history
    
    def runge_kutta(self, y, x, dx, f):
        """ y is the initial value for y
            x is the initial value for x
            dx is the time step in x
            f is derivative of function y(t)
        """
        k1 = dx * f(y, x)
        k2 = dx * f(y + 0.5 * k1, x + 0.5 * dx)
        k3 = dx * f(y + 0.5 * k2, x + 0.5 * dx)
        k4 = dx * f(y + k3, x + dx)
        return y + (k1 + 2 * k2 + 2 * k3 + k4) / 6
